@@ -1,29 +1,20 @@
 package Net::SNMPTrapd;
 
 ########################################################
-#
 # AUTHOR = Michael Vincent
 # www.VinsWorld.com
-#
 ########################################################
 
-require 5.005;
-
 use strict;
-use Exporter;
+use warnings;
 use Convert::ASN1;
 use Socket qw(inet_ntoa AF_INET IPPROTO_TCP);
 
 my $AF_INET6 = eval { Socket::AF_INET6() };
 my $NI_NUMERICHOST = eval { Socket::NI_NUMERICHOST() };
 
-our $VERSION     = '0.10';
-our @ISA         = qw(Exporter);
-our @EXPORT      = qw();
-our %EXPORT_TAGS = (
-                    'all' => [qw()]
-                   );
-our @EXPORT_OK   = (@{$EXPORT_TAGS{'all'}});
+our $VERSION = '0.14';
+our @ISA;
 
 my $HAVE_IO_Socket_IP = 0;
 eval "use IO::Socket::IP -register";
@@ -43,9 +34,67 @@ use constant SNMPTRAPD_RFC_SIZE     => 484;   # RFC limit
 use constant SNMPTRAPD_REC_SIZE     => 1472;  # Recommended size
 use constant SNMPTRAPD_MAX_SIZE     => 65467; # Actual limit (65535 - IP/UDP)
 
-our @TRAPTYPES = qw(COLDSTART WARMSTART LINKDOWN LINKUP AUTHFAIL EGPNEIGHBORLOSS ENTERPRISESPECIFIC);
-our @PDUTYPES  = qw(GetRequest GetNextRequest Response SetRequest Trap GetBulkRequest InformRequest SNMPv2-Trap Report);
+my @TRAPTYPES = qw(COLDSTART WARMSTART LINKDOWN LINKUP AUTHFAIL EGPNEIGHBORLOSS ENTERPRISESPECIFIC);
+my @PDUTYPES  = qw(GetRequest GetNextRequest Response SetRequest Trap GetBulkRequest InformRequest SNMPv2-Trap Report);
 our $LASTERROR;
+
+my $asn = Convert::ASN1->new;
+$asn->prepare("
+    PDU ::= SEQUENCE {
+        version   INTEGER,
+        community STRING,
+        pdu_type  PDUs
+    }
+    PDUs ::= CHOICE {
+        response        Response_PDU,
+        trap            Trap_PDU,
+        inform_request  InformRequest_PDU,
+        snmpv2_trap     SNMPv2_Trap_PDU
+    }
+    Response_PDU      ::= [2] IMPLICIT PDUv2
+    Trap_PDU          ::= [4] IMPLICIT PDUv1
+    InformRequest_PDU ::= [6] IMPLICIT PDUv2
+    SNMPv2_Trap_PDU   ::= [7] IMPLICIT PDUv2
+
+    IPAddress ::= [APPLICATION 0] STRING
+    Counter32 ::= [APPLICATION 1] INTEGER
+    Guage32   ::= [APPLICATION 2] INTEGER
+    TimeTicks ::= [APPLICATION 3] INTEGER
+    Opaque    ::= [APPLICATION 4] STRING
+    Counter64 ::= [APPLICATION 6] INTEGER
+
+    PDUv1 ::= SEQUENCE {
+        ent_oid         OBJECT IDENTIFIER,
+        agent_addr      IPAddress,
+        generic_trap    INTEGER,
+        specific_trap   INTEGER,
+        timeticks       TimeTicks,
+        varbindlist     VARBINDS
+    }
+    PDUv2 ::= SEQUENCE {
+        request_id      INTEGER,
+        error_status    INTEGER,
+        error_index     INTEGER,
+        varbindlist     VARBINDS
+    }
+    VARBINDS ::= SEQUENCE OF SEQUENCE {
+        oid    OBJECT IDENTIFIER,
+        value  CHOICE {
+            integer   INTEGER,
+            string    STRING,
+            oid       OBJECT IDENTIFIER,
+            ipaddr    IPAddress,
+            counter32 Counter32,
+            guage32   Guage32,
+            timeticks TimeTicks,
+            opaque    Opaque,
+            counter64 Counter64,
+            null      NULL
+        }
+    }
+");
+my $snmpasn = $asn->find('PDU');
+
 ########################################################
 # End Variables
 ########################################################
@@ -68,35 +117,36 @@ sub new {
 
     if (@_ == 1) {
         $LASTERROR = "Insufficient number of args - @_";
-        return(undef)
+        return undef
     } else {
         my %cfg = @_;
         for (keys(%cfg)) {
             if (/^-?localport$/i) {
-                $params{'LocalPort'} = $cfg{$_}
+                $params{LocalPort} = $cfg{$_}
             } elsif (/^-?localaddr$/i) {
-                $params{'LocalAddr'} = $cfg{$_}
+                $params{LocalAddr} = $cfg{$_}
             } elsif (/^-?family$/i) {
                  if ($cfg{$_} =~ /^(?:(?:(:?ip)?v?(?:4|6))|${\AF_INET}|$AF_INET6)$/) {
                     if ($cfg{$_} =~ /^(?:(?:(:?ip)?v?4)|${\AF_INET})$/) {
-                        $params{'Family'} = AF_INET
+                        $params{Family} = AF_INET
                     } else {
                         if (!$HAVE_IO_Socket_IP) {
                             $LASTERROR = "IO::Socket::IP required for IPv6";
-                            return(undef)
+                            return undef
                         }
-                        $params{'Family'} = $AF_INET6
+                        $params{Family} = $AF_INET6;
+                        $params{V6Only} = 1
                     }
                 } else {
                     $LASTERROR = "Invalid family - $cfg{$_}";
-                    return(undef)
+                    return undef
                 }
             } elsif (/^-?timeout$/i) {
                 if ($cfg{$_} =~ /^\d+$/) {
-                    $params{'Timeout'} = $cfg{$_}
+                    $params{Timeout} = $cfg{$_}
                 } else {
                     $LASTERROR = "Invalid timeout - $cfg{$_}";
-                    return(undef)
+                    return undef
                 }
             }
         }
@@ -109,7 +159,7 @@ sub new {
                      }, $class
     } else {
         $LASTERROR = "Error opening socket for listener: $@";
-        return(undef)
+        return undef
     }
 }
 
@@ -128,7 +178,7 @@ sub get_trap {
     my $datagramsize = SNMPTRAPD_MAX_SIZE;
     if (@_ == 1) {
         $LASTERROR = "Insufficient number of args: @_";
-        return(undef)
+        return undef
     } else {
         my %args = @_;        
         for (keys(%args)) {
@@ -144,46 +194,48 @@ sub get_trap {
                     $datagramsize = SNMPTRAPD_REC_SIZE
                 } else {
                     $LASTERROR = "Not a valid size: $args{$_}";
-                    return(undef)
+                    return undef
                 }
             # -timeout
             } elsif (/^-?timeout$/i) {
                 if ($args{$_} =~ /^\d+$/) {
-                    $trap->{'Timeout'} = $args{$_}
+                    $trap->{Timeout} = $args{$_}
                 } else {
                     $LASTERROR = "Invalid timeout - $args{$_}";
-                    return(undef)
+                    return undef
                 }
             }
         }
     }
 
-    my $Timeout = $trap->{'Timeout'};
-    my $udpserver = $self->{'_UDPSERVER_'};
+    my $Timeout = $trap->{Timeout};
+    my $udpserver = $self->{_UDPSERVER_};
     my $datagram;
 
-    # vars for IO select
-    my ($rin, $rout, $ein, $eout) = ('', '', '', '');
-    vec($rin, fileno($udpserver), 1) = 1;
+    if ($Timeout != 0) {
+        # vars for IO select
+        my ($rin, $rout, $ein, $eout) = ('', '', '', '');
+        vec($rin, fileno($udpserver), 1) = 1;
 
-    # check if a message is waiting
-    if (select($rout=$rin, undef, $eout=$ein, $Timeout)) {
-        # read the message
-        if ($udpserver->recv($datagram, $datagramsize)) {
-
-            $trap->{'_TRAP_'}{'PeerPort'} = $udpserver->SUPER::peerport;
-            $trap->{'_TRAP_'}{'PeerAddr'} = $udpserver->SUPER::peerhost;
-            $trap->{'_TRAP_'}{'datagram'} = $datagram;
-
-            return bless $trap, $class
-        } else {
-            $LASTERROR = sprintf "Socket RECV error: $!";
-            return(undef)
+        # check if a message is waiting
+        if (! select($rout=$rin, undef, $eout=$ein, $Timeout)) {
+            $LASTERROR = "Timed out waiting for datagram";
+            return(0)
         }
-    } else {
-        $LASTERROR = "Timed out waiting for datagram";
-        return(0)
     }
+
+    # read the message
+    if ($udpserver->recv($datagram, $datagramsize)) {
+
+        $trap->{_TRAP_}{PeerPort} = $udpserver->SUPER::peerport;
+        $trap->{_TRAP_}{PeerAddr} = $udpserver->SUPER::peerhost;
+        $trap->{_TRAP_}{datagram} = $datagram;
+
+        return bless $trap, $class
+    }
+
+    $LASTERROR = sprintf "Socket RECV error: $!";
+    return undef
 }
 
 sub process_trap {
@@ -195,13 +247,13 @@ sub process_trap {
     if (($self eq $class) && ($class eq __PACKAGE__)) {
         my %th;
         $self = \%th;
-        ($self->{'_TRAP_'}{'datagram'}) = @_
+        ($self->{_TRAP_}{datagram}) = @_
     }
     # Net::SNMPTrapd::process_trap($data)
     if ($class ne __PACKAGE__) {
         my %th;
         $self = \%th;
-        ($self->{'_TRAP_'}{'datagram'}) = $class;
+        ($self->{_TRAP_}{datagram}) = $class;
         $class = __PACKAGE__
     }
 
@@ -212,7 +264,7 @@ sub process_trap {
         for (keys(%args)) {
             # -datagram
             if ((/^-?data(?:gram)?$/i) || (/^-?pdu$/i)) {
-                $self->{'_TRAP_'}{'datagram'} = $args{$_}
+                $self->{_TRAP_}{datagram} = $args{$_}
             # -noresponse
             } elsif (/^-?noresponse$/i) {
                 if (($args{$_} =~ /^\d+$/) && ($args{$_} > 0)) {
@@ -222,229 +274,180 @@ sub process_trap {
         }
     }
 
-    my $asn = new Convert::ASN1;
-
-    ### Process first part of PDU (excluding varbinds)
-    my $trap1;
-    $asn->prepare("
-        PDU ::= SEQUENCE {
-            version   INTEGER,
-            community STRING,
-            pdu_type  PDUs
-        }
-        PDUs ::= CHOICE {
-            trap           Trap-PDU,
-            inform-request InformRequest-PDU,
-            snmpv2-trap    SNMPv2-Trap-PDU
-        }
-        Trap-PDU          ::= [4] IMPLICIT PDUv1
-        InformRequest-PDU ::= [6] IMPLICIT PDUv2
-        SNMPv2-Trap-PDU   ::= [7] IMPLICIT PDUv2
-        PDUv1 ::= SEQUENCE {
-            ent_OID         OBJECT IDENTIFIER,
-            agentaddr       [APPLICATION 0] STRING,
-            generic_trap    INTEGER,
-            specific_trap   INTEGER,
-            timeticks       [APPLICATION 3] INTEGER,
-            ber_varbindlist ANY
-        }
-        PDUv2 ::= SEQUENCE {
-            request_ID      INTEGER,
-            error_status    INTEGER,
-            error_index     INTEGER,
-            ber_varbindlist ANY
-        }
-    ");
-    my $found = $asn->find('PDU');
-    if (!defined($trap1 = $found->decode($self->{'_TRAP_'}{'datagram'}))) {
-        $LASTERROR = sprintf "Error decoding PDU - %s", (defined($asn->error) ? $asn->error : "Unknown Convert::ASN1->decode() error.  Consider $class dump()");
-        return(undef)
+    my $trap;
+    if (!defined($trap = $snmpasn->decode($self->{_TRAP_}{datagram}))) {
+        $LASTERROR = sprintf "Error decoding PDU - %s", (defined($snmpasn->error) ? $snmpasn->error : "Unknown Convert::ASN1->decode() error.  Consider $class dump()");
+        return undef
     }
-    #DEBUG: use Data::Dumper; print Dumper \$trap1;
+    #DEBUG: use Data::Dumper; print Dumper \$trap;
 
     # Only understand SNMPv1 (0) and v2c (1)
-    if ($trap1->{'version'} > 1) {
-        $LASTERROR = sprintf "Unrecognized SNMP version - %i", $trap1->{'version'};
-        return(undef)
+    if ($trap->{version} > 1) {
+        $LASTERROR = sprintf "Unrecognized SNMP version - %i", $trap->{version};
+        return undef
     }
 
     # set PDU Type for later use
-    my $pdutype = sprintf "%s", keys(%{$trap1->{'pdu_type'}});
-
-    ### Process varbinds
-    my $trap2;
-    $asn->prepare("
-        varbind SEQUENCE OF SEQUENCE {
-            oid OBJECT IDENTIFIER,
-            choice CHOICE {
-                val_integer   INTEGER,
-                val_string    STRING,
-                val_OID       OBJECT IDENTIFIER,
-                val_IpAddr    [APPLICATION 0] STRING,
-                val_Counter32 [APPLICATION 1] INTEGER,
-                val_Guage32   [APPLICATION 2] INTEGER,
-                val_TimeTicks [APPLICATION 3] INTEGER,
-                val_Opaque    [APPLICATION 4] STRING,
-                val_Counter64 [APPLICATION 6] INTEGER
-            }
-        }
-    ");
-    if (!defined($trap2 = $asn->decode($trap1->{'pdu_type'}->{$pdutype}->{'ber_varbindlist'}))) {
-        $LASTERROR = sprintf "Error decoding varbinds - %s", (defined($asn->error) ? $asn->error : "Unknown Convert::ASN1->decode() error.  Consider $class dump()");
-        return(undef)
-    }
-    #DEBUG: use Data::Dumper; print Dumper \$trap2;
+    my $pdutype = sprintf "%s", keys(%{$trap->{pdu_type}});
 
     ### Assemble decoded trap object
     # Common
-    $self->{'_TRAP_'}{'version'} = $trap1->{'version'};
-    $self->{'_TRAP_'}{'community'} = $trap1->{'community'};
+    $self->{_TRAP_}{version} = $trap->{version};
+    $self->{_TRAP_}{community} = $trap->{community};
     if ($pdutype eq 'trap') {
-        $self->{'_TRAP_'}{'pdu_type'} = 4
+        $self->{_TRAP_}{pdu_type} = 4
     
-    } elsif ($pdutype eq 'inform-request') {
-        $self->{'_TRAP_'}{'pdu_type'} = 6;
+    } elsif ($pdutype eq 'inform_request') {
+        $self->{_TRAP_}{pdu_type} = 6;
 
         # send response for InformRequest
         if ($RESPONSE) {
-            if ((my $r = &_InformRequest_Response(\$self, \$trap1, $pdutype)) ne 'OK') {
+            if ((my $r = _InformRequest_Response(\$self, $trap, $pdutype)) ne 'OK') {
                 $LASTERROR = sprintf "Error sending InformRequest Response - %s", $r;
-                return(undef)
+                return undef
             }
         }
 
-    } elsif ($pdutype eq 'snmpv2-trap') { 
-        $self->{'_TRAP_'}{'pdu_type'} = 7
+    } elsif ($pdutype eq 'snmpv2_trap') { 
+        $self->{_TRAP_}{pdu_type} = 7
     }
 
     # v1
-    if ($trap1->{'version'} == 0) {
-        $self->{'_TRAP_'}{'ent_OID'}       =          $trap1->{'pdu_type'}->{$pdutype}->{'ent_OID'};
-        $self->{'_TRAP_'}{'agentaddr'}     = inetNtoa($trap1->{'pdu_type'}->{$pdutype}->{'agentaddr'});
-        $self->{'_TRAP_'}{'generic_trap'}  =          $trap1->{'pdu_type'}->{$pdutype}->{'generic_trap'};
-        $self->{'_TRAP_'}{'specific_trap'} =          $trap1->{'pdu_type'}->{$pdutype}->{'specific_trap'};
-        $self->{'_TRAP_'}{'timeticks'}     =          $trap1->{'pdu_type'}->{$pdutype}->{'timeticks'};
+    if ($trap->{version} == 0) {
+        $self->{_TRAP_}{ent_oid}       =           $trap->{pdu_type}->{$pdutype}->{ent_oid};
+        $self->{_TRAP_}{agent_addr}    = _inetNtoa($trap->{pdu_type}->{$pdutype}->{agent_addr});
+        $self->{_TRAP_}{generic_trap}  =           $trap->{pdu_type}->{$pdutype}->{generic_trap};
+        $self->{_TRAP_}{specific_trap} =           $trap->{pdu_type}->{$pdutype}->{specific_trap};
+        $self->{_TRAP_}{timeticks}     =           $trap->{pdu_type}->{$pdutype}->{timeticks};
 
     # v2c
-    } elsif ($trap1->{'version'} == 1) {
-        $self->{'_TRAP_'}{'request_ID'}   = $trap1->{'pdu_type'}->{$pdutype}->{'request_ID'};
-        $self->{'_TRAP_'}{'error_status'} = $trap1->{'pdu_type'}->{$pdutype}->{'error_status'};
-        $self->{'_TRAP_'}{'error_index'}  = $trap1->{'pdu_type'}->{$pdutype}->{'error_index'};
+    } elsif ($trap->{version} == 1) {
+        $self->{_TRAP_}{request_id}   = $trap->{pdu_type}->{$pdutype}->{request_id};
+        $self->{_TRAP_}{error_status} = $trap->{pdu_type}->{$pdutype}->{error_status};
+        $self->{_TRAP_}{error_index}  = $trap->{pdu_type}->{$pdutype}->{error_index};
     }
 
     # varbinds
     my @varbinds;
-    for my $i (0..$#{$trap2->{'varbind'}}) {
+    for my $i (0..$#{$trap->{pdu_type}->{$pdutype}->{varbindlist}}) {
         my %oidval;
-        for (keys(%{$trap2->{'varbind'}[$i]->{'choice'}})) {
-            $oidval{$trap2->{'varbind'}[$i]->{'oid'}} = (
-                                                         defined($trap2->{'varbind'}[$i]->{'choice'}{$_}) ? 
-                                                           (($_ eq 'val_IpAddr') ? 
-                                                             inetNtoa($trap2->{'varbind'}[$i]->{'choice'}{$_}) : 
-                                                           $trap2->{'varbind'}[$i]->{'choice'}{$_}) : 
-                                                         ""
-                                                        )
+        for (keys(%{$trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{value}})) {
+            # defined
+            if (defined($trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{value}{$_})) {
+                # special cases:  IP address, null
+                if ($_ eq 'ipaddr') {
+                    $oidval{$trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{oid}} = _inetNtoa($trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{value}{$_})
+                } elsif ($_ eq 'null') {
+                    $oidval{$trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{oid}} = '(NULL)'
+                # no special case:  just assign it
+                } else {
+                    $oidval{$trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{oid}} =           $trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{value}{$_}
+                }
+            # not defined - ""
+            } else {
+                $oidval{$trap->{pdu_type}->{$pdutype}->{varbindlist}[$i]->{oid}} = ""
+            }
         }
         push @varbinds, \%oidval
     }
-    $self->{'_TRAP_'}{'varbinds'} = \@varbinds;
+    $self->{_TRAP_}{varbinds} = \@varbinds;
 
     return bless $self, $class
 }
 
 sub server {
     my $self = shift;
-    return $self->{'_UDPSERVER_'}
+    return $self->{_UDPSERVER_}
 }
 
 sub datagram {
     my ($self, $arg) = @_;
 
     if (defined($arg) && ($arg >= 1)) {
-        return unpack ('H*', $self->{'_TRAP_'}{'datagram'})
+        return unpack ('H*', $self->{_TRAP_}{datagram})
     } else {
-        return $self->{'_TRAP_'}{'datagram'}
+        return $self->{_TRAP_}{datagram}
     }
 }
 
 sub remoteaddr {
     my $self = shift;
-    return $self->{'_TRAP_'}{'PeerAddr'}
+    return $self->{_TRAP_}{PeerAddr}
 }
 
 sub remoteport {
     my $self = shift;
-    return $self->{'_TRAP_'}{'PeerPort'}
+    return $self->{_TRAP_}{PeerPort}
 }
 
 sub version {
     my $self = shift;
-    return $self->{'_TRAP_'}{'version'} + 1
+    return $self->{_TRAP_}{version} + 1
 }
 
 sub community {
     my $self = shift;
-    return $self->{'_TRAP_'}{'community'}
+    return $self->{_TRAP_}{community}
 }
 
 sub pdu_type {
     my ($self, $arg) = @_;
 
     if (defined($arg) && ($arg >= 1)) {
-        return $self->{'_TRAP_'}{'pdu_type'}
+        return $self->{_TRAP_}{pdu_type}
     } else {
-        return $PDUTYPES[$self->{'_TRAP_'}{'pdu_type'}]
+        return $PDUTYPES[$self->{_TRAP_}{pdu_type}]
     }
 }
 
 sub ent_OID {
     my $self = shift;
-    return $self->{'_TRAP_'}{'ent_OID'}
+    return $self->{_TRAP_}{ent_oid}
 }
 
 sub agentaddr {
     my $self = shift;
-    return $self->{'_TRAP_'}{'agentaddr'}
+    return $self->{_TRAP_}{agent_addr}
 }
 
 sub generic_trap {
     my ($self, $arg) = @_;
 
     if (defined($arg) && ($arg >= 1)) {
-        return $self->{'_TRAP_'}{'generic_trap'}
+        return $self->{_TRAP_}{generic_trap}
     } else {
-        return $TRAPTYPES[$self->{'_TRAP_'}{'generic_trap'}]
+        return $TRAPTYPES[$self->{_TRAP_}{generic_trap}]
     }
 }
 
 sub specific_trap {
     my $self = shift;
-    return $self->{'_TRAP_'}{'specific_trap'}
+    return $self->{_TRAP_}{specific_trap}
 }
 
 sub timeticks {
     my $self = shift;
-    return $self->{'_TRAP_'}{'timeticks'}
+    return $self->{_TRAP_}{timeticks}
 }
 
 sub request_ID {
     my $self = shift;
-    return $self->{'_TRAP_'}{'request_ID'}
+    return $self->{_TRAP_}{request_id}
 }
 
 sub error_status {
     my $self = shift;
-    return $self->{'_TRAP_'}{'error_status'}
+    return $self->{_TRAP_}{error_status}
 }
 
 sub error_index {
     my $self = shift;
-    return $self->{'_TRAP_'}{'error_index'}
+    return $self->{_TRAP_}{error_index}
 }
 
 sub varbinds {
     my $self = shift;
-    return $self->{'_TRAP_'}{'varbinds'}
+    return $self->{_TRAP_}{varbinds}
 }
 
 sub error {
@@ -460,22 +463,22 @@ sub dump {
     if (($self eq $class) && ($class eq __PACKAGE__)) {
         my %th;
         $self = \%th;
-        ($self->{'_TRAP_'}{'datagram'}) = @_
+        ($self->{_TRAP_}{datagram}) = @_
     }
     # Net::SNMPTrapd::dump($datagram)
     if ($class ne __PACKAGE__) {
         my %th;
         $self = \%th;
-        ($self->{'_TRAP_'}{'datagram'}) = $class;
+        ($self->{_TRAP_}{datagram}) = $class;
         $class = __PACKAGE__
     }
 
-    if (defined($self->{'_TRAP_'}{'datagram'})) {
-        Convert::ASN1::asn_dump($self->{'_TRAP_'}{'datagram'});
-        Convert::ASN1::asn_hexdump($self->{'_TRAP_'}{'datagram'});
+    if (defined($self->{_TRAP_}{datagram})) {
+        Convert::ASN1::asn_dump($self->{_TRAP_}{datagram});
+        Convert::ASN1::asn_hexdump($self->{_TRAP_}{datagram});
     } else {
         $LASTERROR = "Missing datagram to dump";
-        return(undef)
+        return undef
     }
 
     return 1
@@ -491,69 +494,51 @@ sub dump {
 
 sub _InformRequest_Response {
 
-    my ($self, $trap1, $pdutype) = @_;
+    my ($self, $trap, $pdutype) = @_;
     my $class = ref($$self) || $$self;
 
-    my $asn = new Convert::ASN1;
-    $asn->prepare("
-        PDU ::= SEQUENCE {
-            version   INTEGER,
-            community STRING,
-            pdu_type  [2] IMPLICIT PDUv2
-        }
-        PDUv2 ::= SEQUENCE {
-            request_ID      INTEGER,
-            error_status    INTEGER,
-            error_index     INTEGER,
-            ber_varbindlist ANY
-        }
-    ");
-    my $found = $asn->find('PDU');
-    my $buffer = $found->encode(
-        version      => $$trap1->{'version'},
-        community    => $$trap1->{'community'},
-        pdu_type     => {
-            request_ID      => $$trap1->{'pdu_type'}->{$pdutype}->{'request_ID'},
-            error_status    => $$trap1->{'pdu_type'}->{$pdutype}->{'error_status'},
-            error_index     => $$trap1->{'pdu_type'}->{$pdutype}->{'error_index'},
-            ber_varbindlist => $$trap1->{'pdu_type'}->{$pdutype}->{'ber_varbindlist'}
-        }
-    );
+    # Change from request to response
+    $trap->{pdu_type}{response} = delete $trap->{pdu_type}{inform_request};
+
+    my $buffer = $snmpasn->encode($trap);
     if (!defined($buffer)) {
-        return $asn->error
+        return $snmpasn->error
     }
     #DEBUG print "BUFFER = $buffer\n";
-    if ($$self->{'_TRAP_'}->{'PeerAddr'} eq "") {
+    if (!defined($$self->{_TRAP_}->{PeerAddr}) || ($$self->{_TRAP_}->{PeerAddr} eq "")) {
         return "Peer Addr undefined"
     }
-    if ($$self->{'_TRAP_'}->{'PeerPort'} == 0) {
+    if (!defined($$self->{_TRAP_}->{PeerPort}) || ($$self->{_TRAP_}->{PeerPort} == 0)) {
         return "Peer Port undefined"
     }
 
     my $socket = $class->SUPER::new(
-                                     Proto     => "udp",
-                                     PeerAddr  => $$self->{'_TRAP_'}->{'PeerAddr'},
-                                     PeerPort  => $$self->{'_TRAP_'}->{'PeerPort'},
-                                     # LocalPort should be set, but creates error.
-                                     # Tried setting ReusePort on initial server,
-                                     # but not implemented on Windows.  What to do?
-                                     #LocalPort => SNMPTRAPD_DEFAULT_PORT,
-                                     Family    => $$self->{'Family'}
-                                    ) || return "Can't create Response socket";
+        Proto     => "udp",
+        PeerAddr  => $$self->{_TRAP_}->{PeerAddr},
+        PeerPort  => $$self->{_TRAP_}->{PeerPort},
+        # LocalPort should be set, but creates error.
+        # Tried setting ReusePort on initial server,
+        # but not implemented on Windows.  What to do?
+        #LocalPort => SNMPTRAPD_DEFAULT_PORT,
+        Family    => $$self->{Family}
+    ) || return "Can't create Response socket";
     $socket->send($buffer);
     close $socket;
+
+    # Change back to request from response
+    $trap->{pdu_type}{inform_request} = delete $trap->{pdu_type}{response};
     return ("OK")
 }
 
-sub inetNtoa {
+sub _inetNtoa {
     my ($addr) = @_;
 
     if ($Socket::VERSION >= 1.94) {
         my $name;
         if (length($addr) == 4) {
-            $name = pack_sockaddr_in(0, $addr)
+            $name = Socket::pack_sockaddr_in(0, $addr)
         } else {
-            $name = pack_sockaddr_in6(0, $addr)
+            $name = Socket::pack_sockaddr_in6(0, $addr)
         }
         my ($err, $address) = Socket::getnameinfo($name, $NI_NUMERICHOST);
         if (defined($address)) {
